from .cross import cross


from .maxvol import maxvol
from .maxvol import rect_maxvol


from .tensor import erank
from .tensor import get
from .tensor import getter
from .tensor import mean
from .tensor import mul
from .tensor import norm
from .tensor import rand
from .tensor import recap
from .tensor import truncate


from .utils import confidence
from .utils import get_cdf
